#ifdef _CH_
#pragma package <opencv>
#endif

#include "ipp.h"
#include "cv.h"
#include "highgui.h"
#include <time.h>
#include <math.h>
#include <ctype.h>
#include <stdio.h>

IplImage* image = 0;
IplImage* image_overlay = 0;
IplImage* image_gt= 0;
bool next_frame=true;


IplImage ** img_buffer=NULL;
int BUFFER_SIZE=100;

#define GET_PIXEL(IMG,TYPE,X,Y) ((TYPE)((IMG)->imageData + (IMG)->widthStep*(Y)))[X]
#define GET_PIXEL_COL(IMG,TYPE,NUM_COLS,X,Y,C) ((TYPE)((IMG)->imageData + (IMG)->widthStep*(Y)))[NUM_COLS*X+C]


/******************************************************************************************/
void decorateImage(IplImage *img_in, IplImage *img_gt, IplImage *img_out) {
	cvConvert(img_in,img_out);
	
	float step_x= (img_in->width*1.0f/img_gt->width);
	float step_y= (img_in->height*1.0f/img_gt->height);

	for(int y=0;y<img_out->height;y++)
		for(int x=0;x<img_out->width;x++) {
			int grid_x=(int)floor(x/step_x);
			int grid_y=(int)floor(y/step_y);
			if (GET_PIXEL(img_gt,unsigned char*,grid_x,grid_y)!=0)
				GET_PIXEL_COL(img_out,unsigned char*,3,x,y,2)=255;
		}


	for(int y=0;y<img_gt->height;y++) {
			CvPoint pt1,pt2;
			pt1.x = 1;
			pt1.y = (int)floor(y*step_y);
			pt2.x = img_out->width;
			pt2.y = (int)floor(y*step_y);
			cvLine( img_out, pt1, pt2 , CV_RGB(255,0,0), 1, CV_AA, 0 );
	}

	for(int x=0;x<img_gt->width;x++) {
			CvPoint pt1,pt2;
			pt1.x = (int)floor(x*step_x);
			pt1.y = 1;
			pt2.x = (int)floor(x*step_x);
			pt2.y = img_out->height;
			cvLine( img_out, pt1, pt2 , CV_RGB(255,0,0), 1, CV_AA, 0 );
	}

}


/******************************************************************************************/
void markGroundtruth( int event, int x, int y, int flags, void* param )
{
    if( !image )return;

    if( image->origin )
        y = image->height - y;

    if( event == CV_EVENT_LBUTTONDOWN )
    {
		int gt_x= x / (image->width/image_gt->width);
		int gt_y= y / (image->height/image_gt->height);
		
		unsigned char current_value = GET_PIXEL(image_gt,unsigned char*,gt_x,gt_y);
		GET_PIXEL(image_gt,unsigned char*,gt_x,gt_y) = 255 - current_value;
    }
	else
	if( event == CV_EVENT_RBUTTONDOWN )
    {
		next_frame=true;
	}

}


/******************************************************************************************/
int main(int argc, char** argv) {            
	
    CvCapture* capture = 0;
	int frame_counter=0;
	int frame_cur=0;

	char save_dir[1000]="";\
	char *output_filemask=NULL;

	int frame_start=1;
	int frame_skip=1;

	int grid_rows=10;
	int grid_cols=10;

	bool is_init=false;	
	bool load_gt=false;
	bool load_prev=false;


	if (( argc<= 2) || (strcmp(argv[1],"/?")==0)) {
		printf("Usage groundtruth [source] [grid_rows] [grid_cols]\n");
		printf("\t source      = either 0 for a webcam or a complete path to a video file\n");
		printf("\t frame_start = the frame where is should start from\n");
		printf("\t frame_skip  = how many frames to skip (default should be 1)\n");
		printf("\t grid_rows   = groundtruth grid - number of rows\n");
		printf("\t grid_cols   = groundtruth grid - number of cols\n");
		printf("\t dest		   = filemask (i.e. out-%04d.png)\n");
		printf("\t reload_gt   = if set to 1, it attempts to load existing groundtruth\n");
		printf("-------------------------------------------------------------------\n");
		printf("\t Keys when running:\n");
		printf("\t ' ' - advance to next frame \n");
		printf("\t 'b' - go back to previous frame \n");
		printf("\t 'g' - toggle load groundtruth (useful when reading from buffer) \n");
		return 0;
	}

    if( (argc == 1) || ((argc >= 2) && (strlen(argv[1]) == 1) && isdigit(argv[1][0])))
        capture = cvCaptureFromCAM( (argc >= 2) ? argv[1][0] - '0' : 0 );
    else if( argc >= 2 )
        capture = cvCaptureFromFile( argv[1] );
	
	int p=2;

	p++;
	if (argc >= p) {
		frame_start=atoi(argv[p-1]);
		printf("Frame Start: %d\n",frame_start);
	}

	p++;
	if (argc >= p) {
		frame_skip=atoi(argv[p-1]);
		printf("Frame Skip: %d\n",frame_skip);
	}

	p++;
	if (argc >= p) {
		grid_rows=atoi(argv[p-1]);
		printf("Grid rows: %d\n",grid_rows);
	}

	p++;
	if (argc >= p) {
		grid_cols=atoi(argv[p-1]);
		printf("Grid cols: %d\n",grid_cols);
	}

	p++;
	if (argc >= p) {
		output_filemask=argv[p-1];
		printf("Output Filemask: %s\n",output_filemask);
	}
	else {
		output_filemask=NULL;
	}

	p++;
	if (argc >= p) {
		load_gt=atoi(argv[p-1]);
		printf("Loading Existing Groundtruth: %d\n",load_gt);
	}


	image_gt = cvCreateImage(cvSize(grid_cols,grid_rows), IPL_DEPTH_8U, 1);

	if( capture ) {


		cvNamedWindow( "Image", 1 );
		cvSetMouseCallback( "Image", markGroundtruth, 0 );

        for(;;) {

			if (next_frame) {
				
				if (frame_cur==frame_counter) frame_counter++;
				frame_cur++;

				 //( const CvArr* src, CvArr* dst, int interpolation=CV_INTER_LINEAR )
				bool error=false;
				
				if (frame_cur==frame_counter) {
					for(int skip=0;skip<frame_skip;skip++)	{
						if( !cvGrabFrame( capture )) {
							error=true;
							break;
						}
						image = cvRetrieveFrame( capture );
					}
					if (img_buffer==NULL) {
						// create the image buffer
						img_buffer = (IplImage**)malloc(sizeof(IplImage*)*BUFFER_SIZE);		
						for(int i=0;i<BUFFER_SIZE;i++)
							img_buffer[i] = cvCreateImage(cvSize(image->width,image->height),image->depth,image->nChannels);
					}
					cvCopy(image,img_buffer[frame_cur%BUFFER_SIZE]);

				}
				else {
					image = img_buffer[frame_cur%BUFFER_SIZE];
				}


				if (error) break;
				if (frame_counter < frame_start) {
					printf("\rFrame=%05d ",frame_cur);
					continue;
				}


				if ((load_gt  || load_prev) && output_filemask) {

					char loadFilename[1010];
					sprintf(loadFilename,output_filemask,frame_cur);
					IplImage *img_gt_new = cvLoadImage(loadFilename,0);
					if (img_gt_new) {
						cvReleaseImage(&image_gt);
						image_gt = img_gt_new;
					}
				}
				load_prev=false;
				next_frame=false;
			}

			if( image) {
				if (is_init==false) {
					image_overlay = cvCreateImage(cvSize(image->width,image->height), IPL_DEPTH_8U, 3);
					//image_overlay = (IplImage *) cvClone(image);
					is_init=true;
				}


				decorateImage(image,image_gt,image_overlay);
				cvShowImage( "Image", image_overlay);

				printf("\rFrame=%05d",frame_cur);
			} //if image_large
			else {
				printf("Could not load the image\n");
			}

			char keyPressed = cvWaitKey(1);
			if( keyPressed  == 27 ) break;
			if( (keyPressed  == ' ') || (keyPressed  == 'b')) next_frame=true;

			if (next_frame && output_filemask) {
				char saveFilename[1010];
				sprintf(saveFilename,output_filemask,frame_cur);
				cvSaveImage(saveFilename,image_gt);
			}

			if( keyPressed  == 'b') { 
				load_prev=true;
				if (frame_counter-frame_cur<BUFFER_SIZE-1)
					frame_cur-=2;
				else
					printf("Image buffer limit reached. Please increase the buffer size and re-compile.\n");
			}

			if( keyPressed  == 'g') { 
				load_gt = !load_gt;
				printf("Toggling Load Groundtruth. Now = %d.\n",load_gt);
			}

        }

        cvReleaseCapture( &capture );

		// gotto release the basis and the buffer
		cvDestroyWindow( "Image");
	}


	cvReleaseImage(&image_overlay);
	cvReleaseImage(&image_gt);

	for(int i=0;i<BUFFER_SIZE;i++)
		cvReleaseImage(&img_buffer[i]);
	free(img_buffer);

	printf("Exiting\n");
    return 0;
}